SystemOrganization addCategory: #MutationEngine!

Object subclass: #Mutation
	instanceVariableNames: 'originalState parameters injector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

Mutation subclass: #AstMutation
	instanceVariableNames: 'ast'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!AstMutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:28:13.403'!
applyAst
	
	injector applyAst: self ast toMethod: self method.! !

!AstMutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 18:20:24.774'!
ast
	"Return the AST of the mutated method.
	The ast is lazily obtained from the injector-object."
	
	^ ast ifNil: [ ast :=	injector astOfMethod: self method ]! !

!AstMutation methodsFor: 'visiting AST nodes' stamp: 'ag 5/25/2012 18:18:15.204'!
nodesToVisit
	"Return a collection of symbols, that identify the types of nodes, that the Mutation-instance is interested
	in visiting. The symbol has to be the Xxx in #acceptXxx: for one of the visitor-methods of
	ConfigurableASTVisitor. A message with the same name will be sent to this Mutation-instance when visiting
	such a node."
	
	self subclassResponsibility.! !

!AstMutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:20:52.39'!
visitAst
	
	| visitor |
	visitor := ConfigurableSetASTVisitor new client: self; yourself.
	self nodesToVisit do: [ :symbol | visitor add: (#accept, symbol, #':') asSymbol ].
	visitor visitNode: self ast.! !

AstMutation subclass: #StringLiteralMutation
	instanceVariableNames: 'mutatedNode visitedLiterals'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!StringLiteralMutation class methodsFor: 'parameter creation' stamp: 'ag 5/25/2012 18:03:44.349'!
countStringLiteralsIn: methodAST
	
	| count |
	count := 0.
	ConfigurableBlockASTVisitor new
		on: #acceptLiteralNode: invoke: [ :node |
			node value isString ifTrue: [ count := count + 1 ] ];
		visitNode: methodAST.
	^ count! !

!StringLiteralMutation class methodsFor: 'parameter creation' stamp: 'ag 5/25/2012 18:02:57.11'!
randomParametersFor: methodAST
	
	| stringLiterals |
	stringLiterals := self countStringLiteralsIn: methodAST.
	stringLiterals <= 0 ifTrue: [ ^ nil ]. "No String-literals -> cannot mutate anything there."
	
	^ Dictionary newFrom: {
		#stringReplacementMode -> self stringReplacementModes atRandom.
		#stringLiteralSelection -> stringLiterals atRandom
	}! !

!StringLiteralMutation class methodsFor: 'parameter creation' stamp: 'ag 5/25/2012 18:02:49.401'!
stringReplacementModes
	"See instance-side, #stringReplacementMode"
	
	^ { #shuffle. #randomString. #randomlyReplaceCharacters }! !

!StringLiteralMutation methodsFor: 'visiting AST nodes' stamp: 'ag 5/25/2012 18:11:15.847'!
acceptLiteralNode: node
	
	| selection |
	node value isString ifFalse: [ ^ true ]. "Visit only String-literals."
	visitedLiterals := visitedLiterals + 1.
	selection := self stringLiteralSelection.
	
	((selection isNumber and: [ visitedLiterals = selection ]) or:
	(selection isString and: [ node value = selection ]))
		ifTrue: [ "Found it!!"
			mutatedNode := node ].
	^ mutatedNode isNil! !

!StringLiteralMutation methodsFor: 'visiting AST nodes' stamp: 'ag 5/25/2012 17:00:52.552'!
acceptNode: node
	"Stop the visiting-process, as soon as the correct literal-node is found.
	If the node is never found, we visit the whole tree."
	
	^ mutatedNode isNil! !

!StringLiteralMutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:21:21.712'!
applyMutation
	
	| newLiteralNode |
	mutatedNode := nil. "In case of re-use of this object."
	visitedLiterals := 0.
	self visitAst.
	mutatedNode ifNil: [ ^ self error: 'The configured String-literal was not found!!' ].
	newLiteralNode := mutatedNode clone.
	self mutateNode: newLiteralNode.
	mutatedNode replaceWith: newLiteralNode.
	self applyAst.! !

!StringLiteralMutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:23:21.635'!
mutateNode: newNode
	"The token inside the String-literal-node is modified, to contain the new, mutated, String object."
	
	| token newValue |
	newValue := self mutatedValue: newNode value.
	token := newNode token.
	token
		value: newValue
		start: token start
		stop: token start + newValue size.! !

!StringLiteralMutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:13:04.297'!
mutatedValue: aString
	"Yeah I know... caseOf:... But it's really what we need here!!"
	
	^ self stringReplacementMode
		caseOf: {
			[#randomlyReplaceCharacters] -> [
				String streamContents: [ :stream |
					aString do: [ :char |
						char isSeparator
							ifTrue: [ stream nextPut: char ]
							ifFalse: [ stream nextPut: self randomCharacter ] ] ] ].
			[#randomString] -> [ | size result |
				size := (aString size * 2) atRandom.
				result := String new: size.
				1 to: size do: [ :i | 
					result at: i put: self randomCharacter ].
				result ].
			[#shuffle] -> [
				aString shuffle ]
		}! !

!StringLiteralMutation methodsFor: 'visiting AST nodes' stamp: 'ag 5/25/2012 18:11:01.971'!
nodesToVisit
	
	^ #(
		LiteralNode
		Node
	)! !

!StringLiteralMutation methodsFor: 'private' stamp: 'ag 5/25/2012 17:21:22.461'!
randomCharacter
	"TODO -- only return printable characters?"
	
	^ 255 atRandom asCharacter! !

!StringLiteralMutation methodsFor: 'parameters' stamp: 'ag 5/25/2012 16:20:58.973'!
stringLiteralSelection
	"Parameter type: Number or String, default: 1.
	Defines, how the mutated literal is found in the mutated method.
	A Number value means the n-th String-literal in that method is selected.
	A String value means the first String-literal with that exact value is selected."
	
	^ parameters at: #stringLiteralSelection ifAbsent: [ 1 ]! !

!StringLiteralMutation methodsFor: 'parameters' stamp: 'ag 5/25/2012 17:24:25.381'!
stringReplacementMode
	"Parameter type: Symbol, default: #shuffle.
	Defines, how the String-literal is mutated. This could alternatively be modelled with subclasses of StringLiteralMutation,
	but there are so many imaginable 'modes', which would create 'too many' classes.
	
	Supported modes:
	#shuffle -- The string is shuffled.
	#randomString -- A random length is selected (maximum original length * 2), and populated with random characters.
	#randomlyReplaceCharacters -- Every character (except white space) in the mutated String-literal is replaced by a random other character (except white space).
	"
	
	^ parameters at: #stringReplacementMode ifAbsent: [ #shuffle ]! !

!Mutation class methodsFor: 'instance creation' stamp: 'ag 5/25/2012 14:42:01.678'!
injector: aMutationInjector parameters: aDictionary
	
	^ self basicNew
		injector: aMutationInjector;
		parameters: aDictionary;
		initialize; "Initialize after setting the parameters-dictionary."
		yourself! !

!Mutation class methodsFor: 'parameter creation' stamp: 'ag 5/25/2012 17:51:37.076'!
randomParametersFor: methodAST
	"Subclasses should implement this, and return a Dictionary with random parameters, that are meaningful for 
	instances of those mutation-classes. This is used by a MutationFactory, when creating random mutation-objects.
	All parameters, that are required by the mutation, has to be created. Optional parameters can be left out.
	
	If the subclass decides, that the Mutation cannot be applied on the given method, then nil must be retrned.
	The parameters can be dependant on the mutated method, that is given as AST."
	
	^ Dictionary new! !

!Mutation methodsFor: 'mutating' stamp: 'ag 5/25/2012 18:14:47.995'!
apply
	"Store the current state of the mutated method and apply the mutation.
	#revert has to be called afterwards, to restore the original state.
	To make sure, that the state is really reverted, exceptions between #apply and #revert should be handled carefullly.
	#doMutated: can be used for that."
	
	originalState ifNotNil: 
		[ ^ self error: 'This mutation-object has already been applied!! It is unsafe to apply it again. Invoke #revert first!!' ].
	originalState := injector currentStateOfMethod: self method.
	self applyMutation.! !

!Mutation methodsFor: 'private' stamp: 'ag 5/25/2012 18:18:52.972'!
applyMutation
	"This is the hook for subclasses to actually perform the mutation on the AST, obtained from the injector variable."
	
	self subclassResponsibility.! !

!Mutation methodsFor: 'mutating' stamp: 'ag 5/25/2012 17:28:08.127'!
doMutated: aBlock
	"Evaluated aBlock, with the mutation active. Any exceptions from aBlock are caught to
	make sure, that the mutation is reverted afterwards."
	
	self apply.
	^ aBlock ensure: [ self revert ]! !

!Mutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.444'!
injector
	^ injector! !

!Mutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.475'!
injector: anObject
	injector := anObject! !

!Mutation methodsFor: 'parameters' stamp: 'ag 5/25/2012 14:25:45.021'!
method
	"This parameter indicates the method to be mutated.
	Value: Association (Class -> selector-symbol).
	This parameter is required for all mutations, that change source-code."
	
	^ parameters at: #method! !

!Mutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.387'!
parameters
	^ parameters! !

!Mutation methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:14:56.414'!
parameters: anObject
	parameters := anObject! !

!Mutation methodsFor: 'mutating' stamp: 'ag 5/25/2012 14:42:34.38'!
revert
	"Bring the mutated method to its original state. Only shows any effect, if #apply has been called earlier."
	
	originalState ifNotNil: [
		injector restoreStateOfMethod: self method to: originalState.
		originalState := nil "Mark this mutation-object as reverted, to prevent restoring the state multiple times." ].! !

Object subclass: #MutationFactory
	instanceVariableNames: 'injector scope availableMutations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MutationFactory class methodsFor: 'private' stamp: 'ag 5/25/2012 17:34:54.393'!
availableMutations
	"TODO -- should remove 'abstract' subclasses, if there are any?..."
	
	^ Mutation allSubclasses! !

!MutationFactory class methodsFor: 'instance creation' stamp: 'ag 5/25/2012 17:32:59.191'!
onMethodWrappers
	
	^ self withInjector: MethodWrapperMutationInjector! !

!MutationFactory class methodsFor: 'instance creation' stamp: 'ag 5/25/2012 17:32:46.164'!
onSourcecode
	
	^ self withInjector: SourcecodeMutationInjector! !

!MutationFactory class methodsFor: 'instance creation' stamp: 'ag 5/25/2012 17:32:02.651'!
withInjector: anInjectorClass
	
	^ self basicNew
		injector: anInjectorClass new;
		availableMutations: self availableMutations;
		initialize;
		yourself! !

!MutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.569'!
availableMutations
	^ availableMutations! !

!MutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.6'!
availableMutations: anObject
	availableMutations := anObject! !

!MutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/25/2012 17:53:21.211'!
createMutation: mutationClass
	"Random: the mutation-parameters, the mutated method.
	Explicit: the mutation-class."
	
	^ self
		createMutation: mutationClass
		forMethod: self randomMethod! !

!MutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/25/2012 17:57:52.951'!
createMutation: mutationClass forMethod: method
	"Random: the mutation-parameters.
	Explicit: the mutation-class, the mutated method."
	
	| parameters |
	self flag: #ag.
	"TODO -- if one of the parameters has been chosen randomly, it is not good to just return nil here.
	We should retry and systematically search all combinations of methods // mutations // parameters. until
	an appropriate combination is found, or there really is NO combination available.
	Currently, as work-around, application-code can loop, until a non-nil-result is returned."
	
	parameters := mutationClass randomParametersFor: (injector astOfMethod: method).
	parameters ifNil: [ ^ nil ].
	^ self
		createMutation: mutationClass
		withParameters: parameters
		forMethod: method! !

!MutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/25/2012 15:57:26.936'!
createMutation: mutationClass withParameters: parameterDictionary
	"Random: the mutated method.
	Explicit: the mutation-class, the mutation-parameters."
	
	^ self
		createMutation: mutationClass
		withParameters: parameterDictionary
		forMethod: self randomMethod! !

!MutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/25/2012 15:56:35.319'!
createMutation: mutationClass withParameters: parameterDictionary forMethod: method
	"Random: nothing.
	Explicit: the mutation-class, the mutation-parameters, the mutated method."
	
	^ mutationClass
		injector: injector
		parameters: 
			(parameterDictionary
				at: #method put: method;
				yourself)! !

!MutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.208'!
injector
	^ injector! !

!MutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.487'!
injector: anObject
	injector := anObject! !

!MutationFactory methodsFor: 'private' stamp: 'ag 5/25/2012 15:27:45.983'!
randomMethod
	"Return a random method from the MutationScope."
	
	^ scope randomMethod! !

!MutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/25/2012 16:07:48.271'!
randomMutation
	"Random: the mutation-class, the mutation-parameters, the mutated method.
	Explicit: nothing.
	
	Random mutation-class with explicit mutation-parameters is not supported, because it makes little
	sense to provide parameters for an unknown entity."
	
	^ self randomMutationForMethod: self randomMethod! !

!MutationFactory methodsFor: 'private' stamp: 'ag 5/25/2012 16:08:31.695'!
randomMutationClass
	
	^ availableMutations atRandom! !

!MutationFactory methodsFor: 'mutation creation' stamp: 'ag 5/25/2012 16:07:56.469'!
randomMutationForMethod: method
	"Random: the mutation-class, the mutation-parameters.
	Explicit: the mutated method.
	
	Random mutation-class with explicit mutation-parameters is not supported, because it makes little
	sense to provide parameters for an unknown entity."
	
	^ self
		createMutation: self randomMutationClass
		forMethod: method! !

!MutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.515'!
scope
	^ scope! !

!MutationFactory methodsFor: 'accessing' stamp: 'ag 5/25/2012 14:45:41.543'!
scope: anObject
	scope := anObject! !

MutationFactory subclass: #UniqueMutationFactory
	instanceVariableNames: 'originalScope originalMutationClasses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!UniqueMutationFactory methodsFor: 'initialization' stamp: 'ag 5/25/2012 16:09:06.147'!
initialize
	
	super initialize.
	originalScope := scope deepCopy.
	originalMutationClasses := availableMutations copy.! !

!UniqueMutationFactory methodsFor: 'private' stamp: 'ag 5/25/2012 15:32:27.262'!
randomMethod
	"The scope is fully traversed, avoiding duplicates. Every returned methods is removed from the scope.
	When the scope becomes empty, we replace it with the saved copy.
	The originalScope-copy is taken in initialize, so the scope of this UniqueMutationFactory cannot be changed after instance creation."
	
	scope ifEmpty: [ scope := originalScope deepCopy ].
	^ scope uniqueRandomMethod! !

!UniqueMutationFactory methodsFor: 'private' stamp: 'ag 5/25/2012 16:11:14.238'!
randomMutationClass
	"See the comment in #randomMethod. Same principle, but applied to the mutation-classes.
	This also means, that availableMutations must not be modified after the creation of this instance."
	
	| mutation |
	availableMutations ifEmpty: [ availableMutations := originalMutationClasses copy ].
	mutation := availableMutations atRandom.
	availableMutations remove: mutation.
	^ mutation! !

Object subclass: #MutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MutationInjector commentStamp: 'ag 5/25/2012 14:36:05.548' prior: 0!
A MutationInjector provides services for Mutation objects.
The mutations are handled on AST-level. The Injector provides access to ASTs and methods to apply changed
ASTs to the methods.

In addition, the current 'state' of the methods can be queried and restored later.
This is used to undo/revert the Mutation.
Depending on the Injector-implementation, different objects are used to represent the 'state' of the method. Some implementations also not need any state at all, which might be preferrable in terms of simplicity/safety. These should subclass from StatelessMutationInjector.

Different Injector-implementations (subclasses) apply the modified AST on different levels.
E.g. a SourcecodeMutationInjector recompiles the modified source-code (obtained from the modified AST) inside the class.

Other ideas for Injector-implementations:
 - MethodWrapperMutationInjector - might compile the source-code into a block (or anonymous CompiledMethod object) and use a MethodWrapper, to execute that new version instead of the original. This might make it easier to cleanup afterwards (simply delete the MethodWrapper).
 - BytecodeMutationInjector - compiled the modified source-code and (somehow) set the bytecode directly into the CompiledMethod. Dont know wheterh thats possible, but the idea is that cleaning up is easier, because the source-code of the method has not changed, and we simply have to recompile it.
 - ContextMutationInjector - create a Layer for Mutations, add a layered version of the method, with the modified source-code, after the mutation, simply delete the layer
!

!MutationInjector class methodsFor: 'instance creation' stamp: 'ag 5/25/2012 14:27:16.821'!
default
	"MutationInjector objects are stateless, but create new instances here nonetheless. Avoids class-variables.
	Also, state might be introduced in the future."
	
	^ SourcecodeMutationInjector new! !

!MutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 13:39:41.059'!
applyAst: parseTree toMethod: method
	"parseTree: RBMethodNode, method: class -> selector-symbol.
	Conrete subclasses decide, how the modified AST of a method is applied into the system."

	self subclassResponsibility.! !

!MutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 18:05:54.559'!
astOfMethod: method
	"method: class -> selector-symbol"
	
	^ (method key >> method value) parseTree! !

!MutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 13:43:03.789'!
currentStateOfMethod: method
	"method: class -> selector-symbol.
	Subclasses can return an arbitrary object here, that has later ot be used by #restoreStateOfMethod:to:
	to restore the method to its original state."

	self subclassResponsibility.! !

!MutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 13:43:09.929'!
restoreStateOfMethod: method to: anObject
	"method: class -> selector-symbol.
	anObject: return value of the according #currentStateOfMethod: call. See the comment there."

	self subclassResponsibility.! !

MutationInjector subclass: #SourcecodeMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!SourcecodeMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 18:28:35.864'!
applyAst: parseTree toMethod: method
	"Get the source from the parseTree, and recompile the source inside the class.
	Changes made to the tree try to preseve the structure of the original code."
	
	self applyCode: parseTree newSource toMethod: method.! !

!SourcecodeMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 14:13:17.967'!
applyCode: sourceCode toMethod: method
	"Find the category of the method, and recompile the given code inside the class.
	Trust in the fact, that the code actually compiles to the selector given in the method-parameter."
	
	method key
		compile: sourceCode
		classified:
			(method key organization
				categoryOfElement: method value).! !

!SourcecodeMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 18:08:35.792'!
currentStateOfMethod: method
	"The 'state' of this MutationInjector-implementation is the source-code of the method.'"
	
	^ (method key >> method value) getSource! !

!SourcecodeMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 14:14:01.494'!
restoreStateOfMethod: method to: anObject
	"anObject is the source-code of method, as returned from currentStateOfMethod:.
	Recompile this original code."
	
	self applyCode: anObject toMethod: method.! !

MutationInjector subclass: #StatelessMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

StatelessMutationInjector subclass: #MethodWrapperMutationInjector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MethodWrapperMutationInjector methodsFor: 'AST manipulation' stamp: 'ag 5/25/2012 14:40:27.208'!
applyAst: parseTree toMethod: method
	
	"TODO - Get the source-code from the parseTree, compile it to an anonymous (!!) CompiledMethod, append a MethodWrapper
	to method, that executed the new CompiledMethod instead of the original."
	self flag: #ag.
	self notYetImplemented.! !

!MethodWrapperMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 14:39:46.766'!
restoreStateOfMethod: method to: anObject
	
	"TODO - remove the Mutation-MethodWrapper from the given method. anObject has no meaningfull value here."
	self flag: #ag.
	self notYetImplemented.! !

!StatelessMutationInjector methodsFor: 'state restoring' stamp: 'ag 5/25/2012 14:37:45.545'!
currentStateOfMethod: method
	"The object returned here does not actually matter, just don't return nil, since Mutation uses nil to mark,
	that the state has not yet been stored.
	#restoreStateOfMethod:to: still has to be implemented of subclasses, but the parameter-object is 'meaningless'."
	
	^ #NoState! !

Object subclass: #MutationScope
	instanceVariableNames: 'methods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:02:39.661'!
add: anObject
	"Convenience-method, allowing to add different kinds of objects to the MutationScope."
	
	anObject isBehavior ifTrue: [ ^ self addClasses: { anObject } ].
	anObject isCompiledMethod ifTrue: [ ^ self addCompiledMethod: anObject ].
	anObject class == Association ifTrue: [ ^ self addMethod: anObject ].
	anObject isDictionary ifTrue: [ ^ self addMethodsFromDictionary: anObject ].
	(anObject isCollection and: anObject notEmpty) ifTrue: [ | probe | "Should contain homogenous elements..."
		probe := anObject anyOne.
		probe isBehavior ifTrue: [ ^ self addClasses: anObject ].
		probe isCompiledMethod ifTrue: [ anObject do: [ self addCompiledMethod: probe ] ].
		probe class == Association ifTrue: [ self addMethodsFromDictionary: (Dictionary newFrom: anObject) ] ].
	self error: 'Don''t know how to add this object to this mutation-scope.'.! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:14:32.045'!
addClass: aClass
	
	self addMethods: aClass allMethodSelectors forClass: aClass.! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:12:03.417'!
addClasses: setOfClasses
	
	setOfClasses do: [ :cl | self addClass: cl ].! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 14:57:45.378'!
addCompiledMethod: aCompiledMethod
	
	self addMethod: (aCompiledMethod methodClass -> aCompiledMethod selector).! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 14:59:42.146'!
addMethod: association
	"Class -> selector"
	
	self addMethods: { association value } forClass: association key.! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:14:21.383'!
addMethods: selectors forClass: aClass
	
	| methodSet |
	methodSet := methods at: aClass ifAbsentPut: [ Set new ].
	methodSet addAll: selectors.! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:13:19.294'!
addMethodsFromDictionary: aDictionary
	"Class -> (Set of selector-object). Adds only the specified methods for each class."
	
	aDictionary keysAndValuesDo: [ :class :methodSet |
		self addMethods: methodSet forClass: class ].! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:07:16.4'!
addPackage: aPackageName
	"Add all classes and all methods in the given package."
	
	self addClasses:
		(SystemOrganizer default listAtCategoryNamed: aPackageName).! !

!MutationScope methodsFor: 'organization' stamp: 'ag 5/25/2012 15:22:59.085'!
cleanup
	"Remove all classes from the dictionary, that have no methods associated to them."
	
	| deadClasses |
	deadClasses := Set new.
	methods keysAndValuesDo: [ :class :selectorSet |
		selectorSet ifEmpty: [ deadClasses add: class ] ].
	deadClasses do: [ methods removeKey: deadClasses ].! !

!MutationScope methodsFor: 'initialization' stamp: 'ag 5/25/2012 15:09:17.878'!
initialize
	
	super initialize.
	methods := Dictionary new. "Dictionary from Class to (Set of Symbol)"! !

!MutationScope methodsFor: 'organization' stamp: 'ag 5/25/2012 15:23:08.359'!
isEmpty
	
	^ methods isEmpty! !

!MutationScope methodsFor: 'method selection' stamp: 'ag 5/25/2012 15:23:34.432'!
randomMethod
	"Returns association - Class -> Selector-Symbol. Returns nil, if this scope is empty."
	
	| class |
	self cleanup.
	self isEmpty ifTrue: [ ^ nil ].
	class := methods atRandom. "Returns random key-object."
	^ class -> (methods at: class) atRandom! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:17:51.449'!
removeAll: aDictionary
	"Class -> (Set of selector-symbols). Remove all methods found in that dictionary."
	
	aDictionary keysAndValuesDo: [ :class :selectorSet |
		methods
			at: class
			ifPresent: [ :currentSelectorSet | currentSelectorSet removeAll: selectorSet ]
			ifAbsent: [ "Nothing, continue" ] ].! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:15:19.809'!
removeClass: aClass
	"Removes all methods in that class."
	
	methods remove: aClass ifAbsent: [ ^ self ].! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:25:48.222'!
removeCompiledMethod: aCompiledMethod
	
	self removeMethod: aCompiledMethod methodClass -> aCompiledMethod selector.! !

!MutationScope methodsFor: 'scope manipulation' stamp: 'ag 5/25/2012 15:09:01.58'!
removeMethod: association
	"Class -> selector-symbol. The class is not removed, even if all methods for that class are removed."
	
	methods
		at: association key
		ifPresent: [ :set | set remove: association value ifAbsent: [ ^ self ] ]
		ifAbsent: [ ^ self ].! !

!MutationScope methodsFor: 'method selection' stamp: 'ag 5/25/2012 15:25:24.001'!
uniqueRandomMethod
	"Return a random method and remove that method from this scope.
	This way, the scope can be fully traversed, wihtout hitting duplicates.
	The scope-object can be clonsed before doing that, if it needs to be reused."
	
	| method |
	method := self randomMethod.
	self removeCompiledMethod: method.
	^ method! !

RBProgramNodeVisitor subclass: #ConfigurableASTVisitor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:32:38.48'!
acceptArrayNode: node
	
	(self invokeClient: #acceptArrayNode: with: node)
		ifTrue: [ super acceptArrayNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:33:00.99'!
acceptAssignmentNode: node
	
	(self invokeClient: #acceptAssignmentNode: with: node)
		ifTrue: [ super acceptAssignmentNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:32:53.233'!
acceptBlockNode: node
	
	(self invokeClient: #acceptBlockNode: with: node)
		ifTrue: [ super acceptBlockNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:33:22.235'!
acceptCascadeNode: node
	
	(self invokeClient: #acceptCascadeNode: with: node)
		ifTrue: [ super acceptCascadeNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:33:38.357'!
acceptLiteralArrayNode: node
	
	(self invokeClient: #acceptLiteralArrayNode: with: node)
		ifTrue: [ super acceptLiteralArrayNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:33:55.514'!
acceptLiteralNode: node
	
	(self invokeClient: #acceptLiteralNode: with: node)
		ifTrue: [ super acceptLiteralNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:34:13.561'!
acceptMessageNode: node
	
	(self invokeClient: #acceptMessageNode: with: node)
		ifTrue: [ super acceptMessageNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:34:24.984'!
acceptMethodNode: node
	
	(self invokeClient: #acceptMethodNode: with: node)
		ifTrue: [ super acceptMethodNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:34:45.726'!
acceptPatternBlockNode: node
	
	(self invokeClient: #acceptPatternBlockNode: with: node)
		ifTrue: [ super acceptPatternBlockNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:35:25.724'!
acceptPatternWrapperBlockNode: node
	
	(self invokeClient: #acceptPatternWrapperBlockNode: with: node)
		ifTrue: [ super acceptPatternWrapperBlockNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:35:33.054'!
acceptPragmaNode: node
	
	(self invokeClient: #acceptPragmaNode: with: node)
		ifTrue: [ super acceptPragmaNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:35:40.341'!
acceptReturnNode: node
	
	(self invokeClient: #acceptReturnNode: with: node)
		ifTrue: [ super acceptReturnNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:35:48.678'!
acceptSequenceNode: node
	
	(self invokeClient: #acceptSequenceNode: with: node)
		ifTrue: [ super acceptSequenceNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'visitor-double dispatching' stamp: 'ag 5/25/2012 16:35:57.472'!
acceptVariableNode: node
	
	(self invokeClient: #acceptVariableNode: with: node)
		ifTrue: [ super acceptVariableNode: node ].! !

!ConfigurableASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 17:45:59.3'!
invokeClient: aSymbol with: aNode
	"The return value must be a boolean and indicate, whether we want to continue traversing the tree, or not."
	
	self subclassResponsibility.! !

!ConfigurableASTVisitor methodsFor: 'visiting' stamp: 'ag 5/25/2012 17:40:41.629'!
visitArgument: node
	"Note -- we invoke accept* on the client, instead of visit*"
	
	(self invokeClient: #acceptArgument: with: node)
		ifTrue: [ super visitArgument: node ].! !

!ConfigurableASTVisitor methodsFor: 'visiting' stamp: 'ag 5/25/2012 17:40:44.999'!
visitArguments: node
	"Note -- we invoke accept* on the client, instead of visit*"
	
	(self invokeClient: #acceptArguments: with: node)
		ifTrue: [ super visitArguments: node ].! !

!ConfigurableASTVisitor methodsFor: 'visiting' stamp: 'ag 5/25/2012 17:40:48.098'!
visitNode: node
	"Note -- we invoke accept* on the client, instead of visit*"
	
	(self invokeClient: #acceptNode: with: node)
		ifTrue: [ super visitNode: node ].! !

ConfigurableASTVisitor subclass: #ConfigurableBlockASTVisitor
	instanceVariableNames: 'blocks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!ConfigurableBlockASTVisitor methodsFor: 'initialization' stamp: 'ag 5/25/2012 17:42:24.85'!
initialize
	
	super initialize.
	blocks := Dictionary new.! !

!ConfigurableBlockASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 18:08:00.158'!
invokeClient: aSymbol with: aNode
	
	^ (blocks includesKey: aSymbol)
		ifTrue: [ | result |
			result := (blocks at: aSymbol)
				valueWithEnoughArguments: { aNode. aSymbol }.
			(result isKindOf: Boolean) "The result has to be boolean, allow clients to use blocks, that return arbitrary values."
				ifTrue: [ result ]
				ifFalse: [ true ] ]
		ifFalse: [ true ]! !

!ConfigurableBlockASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 17:47:50.959'!
on: aSymbol invoke: aBlock
	"Configure aBlock to be evaluated, when the visitor visits a node identified by aSymbol.
	aBlock will get up to 2 parameters (the visited node and the symbol identifying it.)"
	
	blocks at: aSymbol put: aBlock.! !

ConfigurableASTVisitor subclass: #ConfigurableClientASTVisitor
	instanceVariableNames: 'client'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

ConfigurableClientASTVisitor subclass: #ConfigurableAutomaticASTVisitor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!ConfigurableAutomaticASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 16:39:59.193'!
clientWantsToVisit: aSymbol
	"This implementation uses inspection to tell whether the client wants to visit a certain node."
	
	^ client respondsTo: aSymbol! !

!ConfigurableClientASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 17:44:37.297'!
client
	^ client! !

!ConfigurableClientASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 17:44:37.58'!
client: anObject
	client := anObject! !

!ConfigurableClientASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 17:44:54.492'!
clientWantsToVisit: aSymbol
	
	self subclassResponsibility.! !

!ConfigurableClientASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 17:44:48.523'!
invokeClient: aSymbol with: node
	
	^ (self clientWantsToVisit: aSymbol)
		ifTrue: [ client perform: aSymbol with: node ]
		ifFalse: [ true ] "Always continue traversing the tree..."! !

ConfigurableClientASTVisitor subclass: #ConfigurableSetASTVisitor
	instanceVariableNames: 'interestingNodes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MutationEngine'!

!ConfigurableSetASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 16:50:04.784'!
add: aSymbol
	"Add a node-type, that the client is interested in visiting."
	
	interestingNodes add: aSymbol.! !

!ConfigurableSetASTVisitor methodsFor: 'private' stamp: 'ag 5/25/2012 16:48:44.058'!
clientWantsToVisit: aSymbol
	"This implementation uses a configurable set of symbols, which the client can fill with the node-types, he wants to visit."
	
	^ interestingNodes includes: aSymbol! !

!ConfigurableSetASTVisitor methodsFor: 'initialization' stamp: 'ag 5/25/2012 16:49:08.67'!
initialize
	
	super initialize.
	interestingNodes := Set new.! !

!ConfigurableSetASTVisitor methodsFor: 'accessing' stamp: 'ag 5/25/2012 16:50:29.223'!
remove: aSymbol
	"Remove a node-type, that the client is interested in visiting."
	
	interestingNodes remove: aSymbol ifAbsent: [ ^ self ].! !
